/************************************************************************
 * MAT 259 - Winter 2020         Final Project                                                
 * Yulei Yuan                                                                       
 * Instructor: George Legrady  TA: Weidi Zhang   
 * Data Source: PSTAT 231                                                           
 *                                                                      
 * Purpose: 3-D visualization on the status of people in different states 
   who voted in the 2016 presidential election.
   
 * Usage: Press buttons on the left to close presentation mode                                                                         
 *        Press space to show status for individual states 
 * Credit to Rodger Luo for the treemap code direction and Weidi's book 
   for understanding the code, thanks!!!
 ***********************************************************************/

import treemap.*; 
import peasy.*; 
import controlP5.*; 

final int NUM_LAYERS = 21;  
                             
final int LAYER_DIST = 80;  

final int TRANS = 100; 

int selectedState = -100; 

boolean presentationMode = true; // true -> automation 
                                  // false -> no automation

PeasyCam cam; 

ControlP5 cp5; 

Table table;
Table stateTable;
int numRows, numColumns;
float[][] dataMatrix;
String[] states;
ArrayList<String> stateNames;
Textarea myTextarea; 
//TreeMap-Related Variables
StateMap[] stateData;
Treemap[] map;

//Variables for drawing rects
ArrayList<ArrayList<PVector>> pointsOne = new ArrayList<ArrayList<PVector>>(NUM_LAYERS);
ArrayList<ArrayList<PVector>> pointsTwo = new ArrayList<ArrayList<PVector>>(NUM_LAYERS);
ArrayList<ArrayList<PVector>> pointsThree = new ArrayList<ArrayList<PVector>>(NUM_LAYERS);
ArrayList<ArrayList<PVector>> pointsFour = new ArrayList<ArrayList<PVector>>(NUM_LAYERS);

//------------------------------------------------------------Setup Section
void setup() {
  size(1280, 720, OPENGL);
  
  //Initial camera and user interface 
  cam = new PeasyCam(this, 2000);
  cp5 = new ControlP5(this);


  //Add innotations at the top of canvas with Control P5 
  Textfield State = cp5.addTextfield("state")
                       .setPosition(160, 10)
                       .setSize(50, 15)
                       .setFocus(true)
                       .setColor(color(0))
                       .setColorBackground(color(255))
                       .setColorActive(color(255))
                       .setColorForeground(color(255))
                       .setCaptionLabel("Please Press Space to See the Change")
                       ;
  State.getCaptionLabel().getStyle().marginTop  = -16; 
  State.getCaptionLabel().getStyle().marginLeft  = 55;
    
  cp5.addCheckBox("checkBox")
     .setPosition(20, 10)
     .setSize(15, 15) 
     .addItem("Presentation Mode", 1)
     .activateAll() //This function makes the check box selected by default
     ;  
  cp5.setAutoDraw(false);
    //add text place for description
  myTextarea = cp5.addTextarea("txt")
                  .setPosition(width-370,30)
                  .setSize(350,280)
                  .setLineHeight(30)
                  .setHeight(35)
                  .setColorBackground(color(40,100))
                  .setColorForeground(color(40,100))
                  .setFont(createFont("SansSerif-24.vlw",16))
                  .setText("Description")
                  .setColor(0)
                  ;
    cp5.setAutoDraw(false);
               
  



  //Put data into Processing table from csv
  table = loadTable("census.csv", "header");
  numRows = table.getRowCount(); 
  numColumns = table.getColumnCount();
  dataMatrix = new float[numColumns-1][numRows];
  states = new String[numRows];
  println("Rows: " + numRows + " Columns: " + numColumns);
  stateTable = loadTable("statenames.csv", "header");
  stateNames = new ArrayList<String>();
  for(int i=0; i<stateTable.getRowCount(); i++){
    if(stateTable.getInt(i,0) % 1 == 0){
      stateNames.add(stateTable.getString(i,1)); 
    }
  }
  
  
  //Put data from the table into a 2D array
  for (int i=0; i<numColumns-1; i++) {
    for (int j=0; j<numRows; j++) {
      dataMatrix[i][j] = table.getFloat(j, i+1);
    }
  }

  for (int i=0; i<numRows; i++) {
    states[i] = table.getString(i, 0);
  }

  //Initialize tree map variables
  stateData = new StateMap[NUM_LAYERS];
  map = new Treemap[NUM_LAYERS];
  for (int i=0; i<NUM_LAYERS; i++) { 
    stateData[i]= new StateMap();
    pointsOne.add(new ArrayList<PVector>());
    pointsTwo.add(new ArrayList<PVector>());
    pointsThree.add(new ArrayList<PVector>());
    pointsFour.add(new ArrayList<PVector>());
  }

  for (int i=0; i<dataMatrix.length; i++) {
    for (int j=0; j<dataMatrix[i].length; j++) {
      stateData[i].addState(states[j], (int)dataMatrix[i][j]);
    }
    stateData[i].finishAdd();
    map[i] = new Treemap(stateData[i], 0, 0, width, height);
  } 

  for (int i=0; i<dataMatrix.length; i++) {
    for (int j=0; j<stateData[i].getItems().length; j++) {
      double x = stateData[i].getItems()[j].getBounds().x;
      double y = stateData[i].getItems()[j].getBounds().y;
      double w = stateData[i].getItems()[j].getBounds().w; 
      double h = stateData[i].getItems()[j].getBounds().h; 
      
      //Based on the central points, caculate coordinates of four corners of the rects 
      PVector pointOne = new PVector((int)x, (int)y, LAYER_DIST*(i-dataMatrix.length/2));
      pointsOne.get(i).add(pointOne);

      PVector pointTwo = new PVector((int)(x+w), (int)y, LAYER_DIST*(i-dataMatrix.length/2));
      pointsTwo.get(i).add(pointTwo);

      PVector pointThree = new PVector((int)(x+w), (int)(y+h), LAYER_DIST*(i-dataMatrix.length/2));
      pointsThree.get(i).add(pointThree);

      PVector pointFour = new PVector((int)x, (int)(y+h), LAYER_DIST*(i-dataMatrix.length/2));
      pointsFour.get(i).add(pointFour);
    }
  }
}

//------------------------------------------------------------Draw Section
void draw() {
  background(220);
  colorMode(HSB); // Change the color mode to HSB
  
  //Set automation in presentation mode 
  if(presentationMode) {
    rotateY(frameCount*0.001);
    if(frameCount % 200 == 0) {
      selectedState += 1;
      if(selectedState == 53) selectedState = -100;
    }
  }

  //The volumetric shapes go through all layers are composed by four side meshes.   
  //The following codes draw four sides of meshes separately.
  
  //Draw Top Meshes
  pushMatrix();
  translate(-width/2, -height/2);
  for (int j=0; j<numRows; j++) {
   for (int i=0; i<pointsOne.size()-1; i++) {
     float x1_1 = pointsOne.get(i).get(j).x;
     float y1_1 = pointsOne.get(i).get(j).y + 0.1;
     float z1_1 = pointsOne.get(i).get(j).z;

     float x2_1 = pointsTwo.get(i).get(j).x;
     float y2_1 = pointsTwo.get(i).get(j).y + 0.1;
     float z2_1 = pointsTwo.get(i).get(j).z;

     float x1_2 = pointsOne.get(i+1).get(j).x;
     float y1_2 = pointsOne.get(i+1).get(j).y + 0.1;
     float z1_2 = pointsOne.get(i+1).get(j).z;

     float x2_2 = pointsTwo.get(i+1).get(j).x;
     float y2_2 = pointsTwo.get(i+1).get(j).y + 0.1;
     float z2_2 = pointsTwo.get(i+1).get(j).z;

     float lineColor = Float.parseFloat(stateData[i].states.keySet().toArray()[j].toString());

     beginShape();  
     noStroke();
     
     //The next a few lines of code controls color of shapes going through layers
     if (lineColor == selectedState) {
       fill(map(lineColor, 1, 52, 0, 255), 255/2, 255,  TRANS*5);
     } else {
       if (selectedState >= 0) noFill();
       else fill(map(lineColor, 1, 52, 0, 255), 255/2, 255/2, TRANS);
     } 
     curveVertex(x1_1, y1_1, z1_1);
     curveVertex(x2_1, y2_1, z2_1);
     curveVertex(x2_2, y2_2, z2_2);
     curveVertex(x1_2, y1_2, z1_2);
     endShape();
   }
  }
  popMatrix();

  //Draw Bottom Meshes 
  pushMatrix();
  translate(-width/2, -height/2);
  for (int j=0; j<numRows; j++) {
   for (int i=0; i<pointsThree.size()-1; i++) {
     float x1_1 = pointsThree.get(i).get(j).x;
     float y1_1 = pointsThree.get(i).get(j).y - 0.1;
     float z1_1 = pointsThree.get(i).get(j).z;

     float x2_1 = pointsFour.get(i).get(j).x;
     float y2_1 = pointsFour.get(i).get(j).y - 0.1;
     float z2_1 = pointsFour.get(i).get(j).z;

     float x1_2 = pointsThree.get(i+1).get(j).x;
     float y1_2 = pointsThree.get(i+1).get(j).y - 0.1;
     float z1_2 = pointsThree.get(i+1).get(j).z;

     float x2_2 = pointsFour.get(i+1).get(j).x;
     float y2_2 = pointsFour.get(i+1).get(j).y - 0.1;
     float z2_2 = pointsFour.get(i+1).get(j).z;

     float lineColor = Float.parseFloat(stateData[i].states.keySet().toArray()[j].toString());     

     beginShape();
     noStroke();
     
     //The next a few lines of code controls color of shapes going through layers
     if (lineColor == selectedState) {
       fill(map(lineColor,1, 52,  0, 255), 255/2, 255,  TRANS*5);
     } else {
       if (selectedState >= 0) noFill();
       else fill(map(lineColor, 1, 52, 0, 255), 255/2, 255/2, TRANS);
     } 
     curveVertex(x1_1, y1_1, z1_1);
     curveVertex(x2_1, y2_1, z2_1);
     curveVertex(x2_2, y2_2, z2_2);
     curveVertex(x1_2, y1_2, z1_2);
     endShape();
   }
  }
  popMatrix();

  //Draw Left Meshes
  pushMatrix();
  translate(-width/2, -height/2);
  for (int j=0; j<numRows; j++) {
   for (int i=0; i<pointsOne.size()-1; i++) {
     float x1_1 = pointsOne.get(i).get(j).x + 0.1;
     float y1_1 = pointsOne.get(i).get(j).y;
     float z1_1 = pointsOne.get(i).get(j).z;

     float x2_1 = pointsFour.get(i).get(j).x + 0.1;
     float y2_1 = pointsFour.get(i).get(j).y;
     float z2_1 = pointsFour.get(i).get(j).z;

     float x1_2 = pointsOne.get(i+1).get(j).x + 0.1;
     float y1_2 = pointsOne.get(i+1).get(j).y;
     float z1_2 = pointsOne.get(i+1).get(j).z;

     float x2_2 = pointsFour.get(i+1).get(j).x + 0.1;
     float y2_2 = pointsFour.get(i+1).get(j).y;
     float z2_2 = pointsFour.get(i+1).get(j).z;

     float lineColor = Float.parseFloat(stateData[i].states.keySet().toArray()[j].toString());     

     beginShape();
     noStroke();
     
     //The next a few lines of code controls color of shapes going through layers
     if (lineColor == selectedState) {
       fill(map(lineColor, 1, 52,  0, 255), 255/2, 255,  TRANS*5);
     } else {
       if (selectedState >= 0) noFill();
       else fill(map(lineColor, 1, 52, 255, 255), 255, 255, TRANS);
     } 
     curveVertex(x1_1, y1_1, z1_1);
     curveVertex(x2_1, y2_1, z2_1);
     curveVertex(x2_2, y2_2, z2_2);
     curveVertex(x1_2, y1_2, z1_2);
     endShape();
   }
  }
  popMatrix();

  //Draw Right Meshes
  pushMatrix();
  translate(-width/2, -height/2);
  for (int j=0; j<numRows; j++) {
   for (int i=0; i<pointsTwo.size()-1; i++) {
     float x1_1 = pointsTwo.get(i).get(j).x - 0.1;
     float y1_1 = pointsTwo.get(i).get(j).y;
     float z1_1 = pointsTwo.get(i).get(j).z;

     float x2_1 = pointsThree.get(i).get(j).x - 0.1;
     float y2_1 = pointsThree.get(i).get(j).y;
     float z2_1 = pointsThree.get(i).get(j).z;

     float x1_2 = pointsTwo.get(i+1).get(j).x - 0.1;
     float y1_2 = pointsTwo.get(i+1).get(j).y;
     float z1_2 = pointsTwo.get(i+1).get(j).z;

     float x2_2 = pointsThree.get(i+1).get(j).x - 0.1;
     float y2_2 = pointsThree.get(i+1).get(j).y;
     float z2_2 = pointsThree.get(i+1).get(j).z;

     float lineColor = Float.parseFloat(stateData[i].states.keySet().toArray()[j].toString());     
     lineColor = map(lineColor, 1, 52, 0, 255);

     beginShape();
     noStroke();
     
     //The next a few lines of code controls color of shapes going through layers
     if (lineColor == selectedState) {
       fill(map(lineColor, 1, 52, 0, 255), 255/2, 255,  TRANS);
     } else {
       if (selectedState >= 0) noFill();
       else fill(map(lineColor, 1, 52, 0, 255), 255/2, 255/2, TRANS);
     } 
     curveVertex(x1_1, y1_1, z1_1);
     curveVertex(x2_1, y2_1, z2_1);
     curveVertex(x2_2, y2_2, z2_2);
     curveVertex(x1_2, y1_2, z1_2);
     endShape();
   }
  }
  popMatrix();

  //Draw tree map layers  
  for (int i=0; i<dataMatrix.length; i++) {
    pushMatrix();
    translate(-width/2, -height/2, LAYER_DIST*(i - dataMatrix.length/2));
    map[i].draw();
    popMatrix();
  }

  //Draw Title
  cam.beginHUD();  //Stopping peasy cam
  fill(150);
  rect(0, 0, width, 35);
  fill(255);
  textAlign(RIGHT);
  if(selectedState >= 1 && selectedState <= 52){
    String stateName = stateNames.get(selectedState);
    stateName = stateName.toUpperCase();
    text(selectedState + " " + stateName, width, 20);
  }
  //textAlign(CENTER);
  //text("2006 U.S. population status by state", 50, -20);
  
  cp5.draw();
  cam.endHUD();
}

//Function for showing each state manually by pressing SPACE
void keyPressed(){
  if(key == ' '){
    if(selectedState < 0) selectedState = 0;
    else selectedState += 1;
    
    if(selectedState == 54) selectedState = -100;
  }
}


public void state(String theText) {
  // automatically receives results from controller input
  println("a textfield event for controller 'input' : "+theText);
  selectedState = int(theText);
}

void checkBox(float[] a) {
  if(a[0] == 1) presentationMode = true;
  else presentationMode = false;
}
